from .conf import conf
from .biotools import sequence_to_record, sequence_to_annotated_record


class StyleObserver:
    """Generic class for observing style-based annotations in sequences.

    The provided subclasses each observe one particular type of DNA sequence
    annotation, such as the highlight color, bold text, underlines, etc.
    """

    def __init__(self):
        pass

    def process_record_features(self, record):
        for feature in record.features:
            self.process_feature(feature)

    def process_feature(self, feature):
        if self.name not in feature.qualifiers:
            return
        value = feature.qualifiers[self.name]
        label = ""
        if "label" in feature.qualifiers:
            label = feature.qualifiers["label"] + "; "
        label += self.name
        if not isinstance(value, bool):
            label += ": " + str(value)
        feature.qualifiers["label"] = label

    def aggregate_features_from_runs(self, runs):
        features = [[None, ""]]
        for run in runs:
            value = self.evaluate(run)
            text = run.text.replace(" ", "")
            if value == features[-1][0]:
                features[-1][1] += text
            else:
                features.append([value, text])
        return features

    def msword_runs_to_record(self, runs, is_protein=False):
        feature_records = [
            (
                sequence_to_annotated_record(
                    text, is_protein=is_protein, **{self.name: val}
                )
                if val
                else sequence_to_record(text, is_protein=is_protein)
            )
            for (val, text) in self.aggregate_features_from_runs(runs)
        ]
        record = feature_records[0]
        if len(feature_records) > 1:
            record = sum(feature_records[1:], record)
        return record


class ColorObserver(StyleObserver):
    """Subclass for color observers."""

    def process_feature(self, feature):
        if self.name not in feature.qualifiers:
            return
        color = feature.qualifiers[self.name]
        for field in ["color", "ApEinfo_revcolor", "ApEinfo_fwdcolor"]:
            feature.qualifiers[field] = color


class CharactersObserver(StyleObserver):
    """Subclass for character-by-character observers."""

    def aggregate_features_from_runs(self, runs):
        features = [[None, ""]]
        text = "".join([r.text for r in runs])
        for character in text:
            value = self.evaluate(character)
            if value == features[-1][0]:
                features[-1][1] += character
            else:
                features.append([value, character])
        return features


class Italic(StyleObserver):
    """Captures italic text."""

    name = "italic"

    def evaluate(self, run):
        """Return whether the run has italic style"""
        return run.italic


class Bold(StyleObserver):
    """Captures bold text."""

    name = "bold"

    def evaluate(self, run):
        """Return whether the run has bold style"""
        return run.bold


class Underline(StyleObserver):
    """Captures underlined text."""

    name = "underline"

    def evaluate(self, run):
        """Return whether the run has underline style"""
        return run.underline


class FontColor(ColorObserver):
    """Captures text with non-black font color."""

    name = "font_color"

    def evaluate(self, run):
        """Return False if no color, else the #ae60bf color."""
        color = str(run.font.color.rgb)
        if color in ["None", "000000"]:
            return False
        else:
            return "#" + color


class HighlightColor(ColorObserver):
    """Captures text with a background-highlighting color."""

    name = "highlight_color"

    def evaluate(self, run):
        """Return False if no background color, else the #ae60bf color."""
        color = run.font.highlight_color
        if color is None:
            return False
        else:
            return conf["color_theme"][color._member_name]


class UpperCase(CharactersObserver):
    """Captures upper-case text."""

    name = "upper_case"

    def evaluate(self, character):
        """Return whether the character is upper"""
        return character == character.upper()


class LowerCase(CharactersObserver):
    """Captures lower-case text."""

    name = "lower_case"

    def evaluate(self, character):  #
        """Return whether the character is lower"""
        return character == character.lower()



from docx import Document

from .Observers import (
    HighlightColor,
    FontColor,
    Bold,
    Italic,
    UpperCase,
    LowerCase,
    Underline,
)
from .biotools import string_is_sequence


class CrazydocParser:
    """Parser to translate MS-docx documents to Biopython records.

    Examples:
    ---------

    >>> parser = CrazydocParser(['highlight_color', 'font_color'])
    >>> biopython_records = parser.parse_doc_file('example.docx')

    Parameters
    ----------

    observers
      A list of either ``crazydoc.Observer`` objects or observer names in
      ``highlight_color``, ``font_color``, ``bold``, ``italic``,
      ``upper_case``, ``lower_case``, ``underline``.
    """

    observers_dict = {
        _class.name: _class()
        for _class in (
            HighlightColor,
            FontColor,
            Bold,
            Italic,
            UpperCase,
            LowerCase,
            Underline,
        )
    }

    def __init__(self, observers):
        self.observers = [
            self.observers_dict[o] if isinstance(o, str) else o for o in observers
        ]

    def _extract_sequence_names_and_runs(self, doc, is_protein=False):
        """Parse the doc, return a list [(sequence_name, sequenceruns), ...]"""
        sequence_name = None
        sequence_paragraphs = []
        reading_sequence = False
        for paragraph in doc.paragraphs:
            stripped = paragraph.text.replace(" ", "")
            if string_is_sequence(stripped, is_protein=is_protein):
                if reading_sequence:
                    sequence_paragraphs[-1][1].append(paragraph)
                else:
                    reading_sequence = True
                    sequence_paragraphs.append((sequence_name, [paragraph]))
            else:
                if reading_sequence:
                    sequence_name = None
                    reading_sequence = False
                if paragraph.text.startswith(">"):
                    sequence_name = paragraph.text[1:].strip()
        return [
            (name, [run for par in paragraphs for run in par.runs])
            for name, paragraphs in sequence_paragraphs
        ]

    def _msword_runs_to_record(self, runs, is_protein=False):
        """Transform a MS Word runs list to a biopython record."""
        records = [
            observer.msword_runs_to_record(runs, is_protein=is_protein)
            for observer in self.observers
        ]
        final_record = records[0]
        record_features = {
            (feature.location.start, feature.location.end): feature
            for feature in final_record.features
        }
        for record in records[1:]:
            for feature in record.features:
                location = feature.location.start, feature.location.end
                if location in record_features:
                    qualifiers = record_features[location].qualifiers
                    qualifiers.update(feature.qualifiers)
                else:
                    final_record.features.append(feature)
                    record_features[location] = feature
        return final_record

    def parse_doc_file(self, filepath=None, doc=None, is_protein=False):
        """Return a list of records, 1 for each sequence contained in the docx.

        Parameters
        ----------

        filepath
          A path to a docx file.

        doc
          A python-docx Document object, which can be provided instead of the
          file path.

        is_protein
          True if the sequences are protein sequences (default: False).
        """
        if doc is None:
            doc = Document(filepath)
        records = []
        for name, runs in self._extract_sequence_names_and_runs(
            doc, is_protein=is_protein
        ):
            record = self._msword_runs_to_record(runs, is_protein=is_protein)
            if name is not None:
                record.id = name
                record.name = name.replace(" ", "_")
            for observer in self.observers:
                observer.process_record_features(record)
            records.append(record)
        return records



try:
    from dna_features_viewer import BiopythonTranslator

    class CrazydocSketcher(BiopythonTranslator):
        """Custom sequence plotter for records from crazydoc.

        Examples:
        ---------
        >>> # plot and export the schema of records from crazydoc
        >>> translator = CrazydocGraphicTranslator()
        >>> for record in records:
        >>>     ax, _ = translator.translate_record(record).plot();
        >>>     ax.figure.savefig(record.id + ".png")
        """

        default_feature_color = "#ffffff"

        def compute_feature_label(self, feature):
            return feature.qualifiers.get("label", None)


except ImportError:

    class CrazydocSketcher:
        """Unavailable class: install DnaFeaturesViewer.

        Try: pip install dna_features_viewer
        """

        def __init__(self):
            raise ImportError(
                "The graphic translator requires "
                "DnaFeaturesViewer to be installed.\n"
                "(try: pip install dna_features_viewer)."
            )



from crazydoc import CrazydocParser, CrazydocSketcher, records_to_genbank
import os

parser = CrazydocParser(["highlight_color", "bold", "underline"])
biopython_records = parser.parse_doc_file("example.docx")

output_dir = "examples_outputs"
if not os.path.exists(output_dir):
    os.mkdir(output_dir)
records_to_genbank(biopython_records, path=output_dir)

sketcher = CrazydocSketcher()
for record in biopython_records:
    sketch = sketcher.translate_record(record)
    ax, _ = sketch.plot()
    ax.set_title(record.id)
    filepath = os.path.join("examples_outputs", "%s.png" % record.id)
    ax.figure.savefig(filepath, bbox_inches="tight")

print("The genbanks and images files are now in folder examples_ouputs/")



import os

from copy import deepcopy
from Bio.Seq import Seq
from Bio import SeqIO
from Bio.SeqRecord import SeqRecord

try:
    # Biopython <1.78
    from Bio.Alphabet import DNAAlphabet
    from Bio.Alphabet import generic_protein

    has_dna_alphabet = True
except ImportError:
    # Biopython >=1.78
    has_dna_alphabet = False
from Bio.SeqFeature import SeqFeature, FeatureLocation


def annotate_record(
    seqrecord, location="full", feature_type="misc_feature", margin=0, **qualifiers
):
    """Add a feature to a Biopython SeqRecord.

    Parameters
    ----------

    seqrecord
      The Biopython SeqRecord to be annotated.

    location
      Either (start, end) or (start, end, strand). (strand defaults to +1).

    feature_type
      The type associated with the feature.

    margin
      Number of extra bases added on each side of the given location.

    qualifiers
      Dictionary that will be the Biopython feature's `qualifiers` attribute.
    """
    if location == "full":
        location = (margin, len(seqrecord) - margin)

    strand = location[2] if len(location) == 3 else 1
    seqrecord.features.append(
        SeqFeature(
            FeatureLocation(location[0], location[1], strand),
            qualifiers=qualifiers,
            type=feature_type,
        )
    )


def sequence_to_record(
    sequence, id="<unknown id>", name="<unknown name>", is_protein=False
):
    """Return a SeqRecord of the sequence, ready to be Genbanked."""
    if has_dna_alphabet:  # Biopython <1.78
        if is_protein:
            seq = Seq(sequence, alphabet=generic_protein)
        else:
            seq = Seq(sequence, alphabet=DNAAlphabet())
    else:
        seq = Seq(sequence)

    seqrecord = SeqRecord(seq, id=id, name=name)
    if is_protein:
        seqrecord.annotations["molecule_type"] = "protein"
    else:
        seqrecord.annotations["molecule_type"] = "DNA"

    return seqrecord


def sequence_to_annotated_record(
    sequence,
    id="<unknown id>",
    name="<unknown name>",
    feature_type="misc_feature",
    is_protein=False,
    **qualifiers
):
    record = sequence_to_record(sequence, is_protein=is_protein)
    annotate_record(record, feature_type=feature_type, **qualifiers)
    return record


genetic_code = "ABCDGHKMNRSTVWY"
nucleotides_set = set(genetic_code + genetic_code.lower())
protein_code = "ACDEFGHIKLMNPQRSTVWYX"
aminoacids_set = set(protein_code + protein_code.lower())


def string_is_sequence(string, is_protein=False):
    if is_protein:
        return len(string) and (set(string) <= aminoacids_set)
    else:
        return len(string) and (set(string) <= nucleotides_set)


def records_to_genbank(records, path=".", extension=None, is_protein=False):
    if extension is None:
        if is_protein:
            extension = ".gp"
        else:
            extension = ".gbk"
    name_duplicates = {}
    for record in records:
        name = record.id.replace("/", "-")
        if name in name_duplicates:
            name_duplicates[name] += 1
            name = "%s_%03d" % (name, name_duplicates[name])
        out_record = deepcopy(record)
        out_record.name = record.name[:20]
        SeqIO.write(out_record, os.path.join(path, name + extension), "genbank")



""" dna_sequencing_viewer/__init__.py """

# __all__ = []

from .conf import conf
from .Observers import StyleObserver
from .CrazydocParser import CrazydocParser
from .plots import CrazydocSketcher
from .biotools import records_to_genbank
from .version import __version__
